/*
 * linux/drivers/mtd/maps/mtd-armadillo.c
 *
 * Copyright (C) 2001-2003 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo map driver"
#define LICENSE "GPL v2"

__u8 armadillo_read8(struct map_info *map, unsigned long ofs)
{
	DEBUG(printk ("armadillo_read8: 0x%x -> [0x%x]\n", ofs, __raw_readb(map->map_priv_1 + ofs)));
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 armadillo_read16(struct map_info *map, unsigned long ofs)
{
	DEBUG(printk ("armadillo_read16: 0x%x -> [0x%x]\n", ofs, __raw_readw(map->map_priv_1 + ofs)));
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 armadillo_read32(struct map_info *map, unsigned long ofs)
{
	DEBUG(printk ("armadillo_read32: 0x%x -> [0x%x]\n", ofs, __raw_readl(map->map_priv_1 + ofs)));
	return __raw_readl(map->map_priv_1 + ofs);
}

void armadillo_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	DEBUG(printk ("armadillo_write8: [0x%x] -> 0x%x\n", d, adr));
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void armadillo_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	DEBUG(printk ("armadillo_write16: [0x%x] -> 0x%x\n", d, adr));
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void armadillo_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	DEBUG(printk ("armadillo_write32: [0x%x] -> 0x%x\n", d, adr));
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void armadillo_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	DEBUG(printk ("armadillo_copy_from: 0x%x@0x%x -> 0x%x\n", len, from, to));
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void armadillo_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	DEBUG(printk ("armadillo_copy_to: 0x%x@0x%x -> 0x%x\n", len, from, to));
	while(len) {
		__raw_writeb(*(unsigned char *) from, map->map_priv_1 + to);
		from++;
		to++;
		len--;
	}
}


#ifdef CONFIG_MTD_ARMADILLO_FLASH
#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *flash_mtd;

struct map_info armadillo_flash_map = {
	name: "flash",
	size: FLASH_SIZE,
	buswidth: FLASH_WIDTH,
	read8: armadillo_read8,
	read16: armadillo_read16,
	read32: armadillo_read32,
	copy_from: armadillo_copy_from,
	write8: armadillo_write8,
	write16: armadillo_write16,
	write32: armadillo_write32,
	copy_to: armadillo_copy_to,
};

struct mtd_partition armadillo_partitions[] = {
	{
		name: "Flash/Bootloader",
		size: 0x8000,
		offset: 0,
#ifndef CONFIG_MTD_ARMADILLO_WRITEABLEALL
		mask_flags: MTD_WRITEABLE, /* force read-only */
#endif
	},
	{
		name: "Flash/Free",
		size: 0x8000,
		offset: MTDPART_OFS_APPEND,
	},
	{
		name: "Flash/KernelImage",
		size: 0x170000,
		offset: MTDPART_OFS_APPEND,
#ifndef CONFIG_MTD_ARMADILLO_WRITEABLEALL
		mask_flags: MTD_WRITEABLE, /* force read-only */
#endif
	},
	{
		name: "Flash/Initrd",
		size: MTDPART_SIZ_FULL,
		offset: MTDPART_OFS_APPEND,
#ifndef CONFIG_MTD_ARMADILLO_WRITEABLEALL
		mask_flags: MTD_WRITEABLE, /* force read-only */
#endif
	},
};

struct resource armadillo_flash_resource = {
	name:   "Flash",
	start:  FLASH_START,
	end:    FLASH_START + FLASH_SIZE - 1,
	flags:  IORESOURCE_IO | IORESOURCE_BUSY,
};

static int __init init_armadillo_flash (void)
{
	struct mtd_partition *parts;
	int nb_parts = 0;
	char *part_type;
	int err;
	
	part_type = "static";
	parts = armadillo_partitions;
	nb_parts = NB_OF(armadillo_partitions);
	
	if (request_resource (&ioport_resource, &armadillo_flash_resource)) {
		printk(KERN_NOTICE "Failed to reserve Armadillo FLASH space\n");
		err = -EBUSY;
		goto out;
	}
	
	armadillo_flash_map.map_priv_1 = (unsigned long)ioremap(FLASH_START, FLASH_SIZE);
	if (!armadillo_flash_map.map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap Armadillo FLASH space\n");
		err = -EIO;
		goto out_resource;
	}

	flash_mtd = do_map_probe("cfi_probe", &armadillo_flash_map);
	if (!flash_mtd) {
		flash_mtd = do_map_probe("map_rom", &armadillo_flash_map);
	}
	if (!flash_mtd) {
		printk("FLASH probe failed\n");
		err = -ENXIO;
		goto out_ioremap;
	}

	flash_mtd->module = THIS_MODULE;
	//flash_mtd->erasesize = 0x2000;
	
	if (add_mtd_partitions(flash_mtd, parts, nb_parts)) {
		printk("FLASH partitions addition failed\n");
		err = -ENOMEM;
		goto out_probe;
	}
		
	return 0;

out_probe:
	map_destroy(flash_mtd);
	flash_mtd = 0;
out_ioremap:
	iounmap((void *)armadillo_flash_map.map_priv_1);
out_resource:
	release_resource (&armadillo_flash_resource);
out:
	return err;
}
#endif





static struct mtd_info *sram_mtd;

struct map_info armadillo_sram_map = {
	name: "SRAM",
	size: SRAM_SIZE,
	buswidth: SRAM_WIDTH,
	read8: armadillo_read8,
	read16: armadillo_read16,
	read32: armadillo_read32,
	copy_from: armadillo_copy_from,
	write8: armadillo_write8,
	write16: armadillo_write16,
	write32: armadillo_write32,
	copy_to: armadillo_copy_to
};

EXPORT_SYMBOL(armadillo_sram_map);

struct resource armadillo_sram_resource = {
	name:   "SRAM",
	start:  SRAM_START,
	end:    SRAM_START + SRAM_SIZE - 1,
	flags:  IORESOURCE_IO | IORESOURCE_BUSY,
};

static int __init init_armadillo_sram (void)
{
	int err;
	
	if (request_resource (&ioport_resource, &armadillo_sram_resource)) {
		printk(KERN_NOTICE "Failed to reserve Armadillo SRAM space\n");
		err = -EBUSY;
		goto out;
	}
	
	armadillo_sram_map.map_priv_1 = (unsigned long)ioremap(SRAM_START, SRAM_SIZE);
	if (!armadillo_sram_map.map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap Armadillo SRAM space\n");
		err = -EIO;
		goto out_resource;
	}

	sram_mtd = do_map_probe("map_ram", &armadillo_sram_map);
	if (!sram_mtd) {
		printk("SRAM probe failed\n");
		err = -ENXIO;
		goto out_ioremap;
	}

	sram_mtd->module = THIS_MODULE;
	//sram_mtd->erasesize = 16;
	
	if (add_mtd_device(sram_mtd)) {
		printk("SRAM device addition failed\n");
		err = -ENOMEM;
		goto out_probe;
	}
		
	return 0;

out_probe:
	map_destroy(sram_mtd);
	sram_mtd = 0;
out_ioremap:
	iounmap((void *)armadillo_sram_map.map_priv_1);
out_resource:
	release_resource (&armadillo_sram_resource);
out:
	return err;
}







static struct mtd_info *bootrom_mtd;

struct map_info armadillo_bootrom_map = {
	name: "BootROM",
	size: BOOTROM_SIZE,
	buswidth: BOOTROM_WIDTH,
	read8: armadillo_read8,
	read16: armadillo_read16,
	read32: armadillo_read32,
	copy_from: armadillo_copy_from,
};

struct resource armadillo_bootrom_resource = {
	name:   "BootROM",
	start:  BOOTROM_START,
	end:    BOOTROM_START + BOOTROM_SIZE - 1,
	flags:  IORESOURCE_IO | IORESOURCE_BUSY,
};

static int __init init_armadillo_bootrom (void)
{
	int err;
	
	if (request_resource (&ioport_resource, &armadillo_bootrom_resource)) {
		printk(KERN_NOTICE "Failed to reserve Armadillo BOOTROM space\n");
		err = -EBUSY;
		goto out;
	}
	
	armadillo_bootrom_map.map_priv_1 = (unsigned long)ioremap(BOOTROM_START, BOOTROM_SIZE);
	if (!armadillo_bootrom_map.map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap Armadillo BootROM space\n");
		err = -EIO;
		goto out_resource;
	}

	bootrom_mtd = do_map_probe("map_rom", &armadillo_bootrom_map);
	if (!bootrom_mtd) {
		printk("BootROM probe failed\n");
		err = -ENXIO;
		goto out_ioremap;
	}

	bootrom_mtd->module = THIS_MODULE;
	//bootrom_mtd->erasesize = 0x10000;
	
	if (add_mtd_device(bootrom_mtd)) {
		printk("BootROM device addition failed\n");
		err = -ENOMEM;
		goto out_probe;
	}
		
	return 0;

out_probe:
	map_destroy(bootrom_mtd);
	bootrom_mtd = 0;
out_ioremap:
	iounmap((void *)armadillo_bootrom_map.map_priv_1);
out_resource:
	release_resource (&armadillo_bootrom_resource);
out:
	return err;
}





static int __init init_armadillo_maps(void)
{

       	printk(KERN_INFO "Armadillo MTD mappings:\n  SRAM 0x%x at 0x%x\n  BootROM 0x%x at 0x%x\n"
#ifdef CONFIG_MTD_ARMADILLO_FLASH
	       "  Flash 0x%x at 0x%x\n"
#endif
	       , 
	       SRAM_SIZE, SRAM_START, BOOTROM_SIZE, BOOTROM_START
#ifdef CONFIG_MTD_ARMADILLO_FLASH
	       , FLASH_SIZE, FLASH_START
#endif
	       );

	init_armadillo_sram();
	init_armadillo_bootrom();
#ifdef CONFIG_MTD_ARMADILLO_FLASH
	init_armadillo_flash();
#endif
	
	return 0;
}
	

static void __exit cleanup_armadillo_maps(void)
{
	if (sram_mtd) {
		del_mtd_device(sram_mtd);
		map_destroy(sram_mtd);
		iounmap((void *)armadillo_sram_map.map_priv_1);
		release_resource (&armadillo_sram_resource);
	}
	
	if (bootrom_mtd) {
		del_mtd_device(bootrom_mtd);
		map_destroy(bootrom_mtd);
		iounmap((void *)armadillo_bootrom_map.map_priv_1);
		release_resource (&armadillo_bootrom_resource);
	}
	
#ifdef CONFIG_MTD_ARMADILLO_FLASH
	if (flash_mtd) {
		del_mtd_partitions(flash_mtd);
		map_destroy(flash_mtd);
		iounmap((void *)armadillo_flash_map.map_priv_1);
		release_resource (&armadillo_flash_resource);
	}
#endif
}

module_init(init_armadillo_maps);
module_exit(cleanup_armadillo_maps);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
MODULE_DESCRIPTION("Armadillo map driver");
MODULE_LICENSE("GPL");
