/*
 * linux/drivers/ssi/ssi-max149.c
 *
 * Copyright (C) 2001-2003 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include "ssi_bus.h"
#include "ssi_dev.h"

#include <linux/ssi-max149.h>

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "ssi MAX149 driver"
#define LICENSE "GPL v2"

#define CHRDEV   "ssi-max149"
#define VERSION  CHRDEV ": " DESCRIPTION ", (C) 2001-2003 " AUTHOR

#define CMD_BASE 0x82             // START | InternalClockMode

extern struct ssi_bus clps711x_ssi1_bus;

static volatile u_int uni = 0x08; // 0x00:bipolar 0x08:unipolar
static          int   recvbuf[1];
static volatile u_int rxed;

static void max149_rcv(struct ssi_dev *dev, u_int data)
{
  recvbuf[0] = (data >> 5) & 0x3ff; // 10bit
  if (!uni && (recvbuf[0] & 0x200)) {
    //bipolar
    recvbuf[0] = (int)(recvbuf[0] | 0xfffffc00);
  }
  rxed = 1;
}

extern struct ssi_dev max149_dev;

static int max149_dev_init(struct ssi_dev *dev)
{
  return 0;
}

static struct ssi_dev max149_dev = {
  name:	    "MAX149",
  id:	    0,
  proto:    SSI_SPI,
  cfglen:   8,
  framelen: 24,
  clkpol:   0,
  clkfreq:  2000000,
  rcv:	    max149_rcv,
  init:	    max149_dev_init,
};

static ssize_t max149_read (struct file *file, char *buf, size_t count,
			    loff_t *ppos)
{
  unsigned int  minor   = MINOR(file->f_dentry->d_inode->i_rdev);
  u_int         command = CMD_BASE | uni;
  unsigned long timeout;
  
  if (!(minor & MAX149_MINORD0_1)) {
    command |= 0x04;
  }
  switch (minor & ~MAX149_MINORD0_1) {
  case 0:
    //command |= 0x00;
    break;
  case 1:
    command |= 0x40;
    break;
  case 2:
    command |= 0x10;
    break;
  case 3:
    command |= 0x50;
    break;
  case 4:
    command |= 0x20;
    break;
  case 5:
    command |= 0x60;
    break;
  case 6:
    command |= 0x30;
    break;
  case 7:
    command |= 0x70;
    break;
  default:
    return -ENODEV;
  }
  
  ssi_select_device (max149_dev.bus, &max149_dev);
  
  rxed = 0;
  ssi_transmit_data (&max149_dev, command);
  timeout = jiffies + 2;
  while (!rxed) {
    if (jiffies >= timeout) {
      // Timed out, retry
      ssi_transmit_data (&max149_dev, command);
      timeout = jiffies + 2;
      while (!rxed && jiffies < timeout) {
	schedule();
      }
      break;
    }
    schedule();
  }
  
  ssi_select_device (max149_dev.bus, NULL);
  
  if (!rxed) {
    return -EFAULT;
  }
  
  if (copy_to_user (buf, recvbuf, sizeof(int))) {
    return -EFAULT;
  }
  
  return sizeof(int);
}

static int max149_ioctl (struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
  switch (cmd) {
  case MAX149_IOCTL_BIP:
    uni = 0x00;
    return 0;
    
  case MAX149_IOCTL_UNI:
    uni = 0x08;
    return 0;
  }
  
  return -EINVAL;
}

static struct file_operations max149_fops = {
  owner: THIS_MODULE,
  read:  max149_read,
  ioctl: max149_ioctl,
};

static int __init max149_init(void)
{
  int ret;
  
  ret = ssi_register_device (&clps711x_ssi1_bus, &max149_dev);
  if (ret) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s device.\n",
	    max149_dev.name);
    return ret;
  }
  
  if(register_chrdev (MAX149_MAJOR, CHRDEV, &max149_fops)) {
    printk (KERN_WARNING CHRDEV ": Unable to get major %d.\n", MAX149_MAJOR);
    ssi_unregister_device (&max149_dev);
    return -EIO;
  }
  
  printk (KERN_INFO VERSION "\n");
  return 0;
}

static void __exit max149_exit(void)
{
  unregister_chrdev (MAX149_MAJOR, CHRDEV);
  
  ssi_unregister_device (&max149_dev);
}

module_init(max149_init);
module_exit(max149_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
